#include <net.h>

ClientOrServer *client_or_server;

Connection *connection_new(void) {

  Connection *new = (Connection*)g_malloc0(sizeof(Connection));
  
  return new;

}

void connection_destroy(Connection *c) {

  if(c == NULL) return;
  
  connection_close(c);
  
  if(c->ip != NULL) {
    g_free(c->ip);
    c->ip = NULL;
  }
  
  g_free(c);
  c = NULL;

}

gboolean connection_open(Connection *c) {

  gboolean rv = FALSE;

  g_assert(c != NULL);
  g_assert(c->ip != NULL);
  g_assert(c->port != 0);

  c->socket = gnet_tcp_socket_connect(c->ip,c->port);
  
  if(c->socket != NULL) {
    c->iochannel = gnet_tcp_socket_get_iochannel(c->socket);
    g_assert(c->iochannel != NULL);
    c->watch_tag = g_io_add_watch(c->iochannel,G_IO_IN,handling_iofunc,c);
    rv = TRUE;
  } 

  return rv;

}

void connection_close(Connection *c) {

  if(c == NULL) return;

  if(c->iochannel != NULL) {
    g_io_channel_close(c->iochannel);
    g_io_channel_unref(c->iochannel);
    c->iochannel = NULL;
  }
  if(c->socket != NULL) {
    gnet_tcp_socket_delete(c->socket);    
  }
  /* you have to free your ip thingy yourself
   * when you change it... or by calling connection_destroy.
   */
  if(c->watch_tag != 0) {
    g_source_remove(c->watch_tag);
  }
  
}

void connection_listen(Connection *c,gint port) {

  g_print("%s: listen\n",__FUNCTION__);

  if(c == NULL) c = connection_new();
  g_assert(c != NULL);

  c->socket = gnet_tcp_socket_server_new(port);
  g_assert(c->socket != NULL);

  c->ip = gnet_inetaddr_gethostname();
  c->port = port;

  c->iochannel = gnet_tcp_socket_get_iochannel(c->socket);
  g_io_add_watch(c->iochannel,ANY_IO_CONDITION,accepting_iofunc,c);

}


gboolean accepting_iofunc(GIOChannel* iochannel,GIOCondition condition, 
			  gpointer data) {

  Connection *listener = (Connection*)data;

  g_assert(listener->socket != NULL);
  
  switch(condition) {
  case G_IO_IN:
    {
      Connection *incoming = connection_new();
      GInetAddr *address;

      incoming->socket = gnet_tcp_socket_server_accept(listener->socket);
      g_assert(incoming->socket != NULL);

      incoming->iochannel = gnet_tcp_socket_get_iochannel(incoming->socket);
      g_assert(incoming->iochannel != NULL);

      address = 
	gnet_tcp_socket_get_inetaddr((const GTcpSocket*)incoming->socket);
      g_assert(address != NULL);
      incoming->ip = gnet_inetaddr_get_name(address);
      incoming->port = gnet_inetaddr_get_port(address);
      gnet_inetaddr_delete(address);

      g_print("%s: new client (%s %d)\n",
	      __FUNCTION__,incoming->ip,incoming->port);

      incoming->watch_tag = g_io_add_watch(incoming->iochannel,
	                                   G_IO_IN | G_IO_HUP | G_IO_ERR,
		                           handling_iofunc,incoming);
    } 
    break;
  default: 
    {
      gchar *str = g_strdup_printf("%s: unknown condition %d",
				   __FUNCTION__,condition);
      current_error_message_set(str);
      current_error_message_show();
      g_free(str);
    } 
    break;
  }
  return TRUE;

}

gboolean handling_iofunc(GIOChannel* iochannel,GIOCondition condition,
			 gpointer data) {

  Connection *incoming = (Connection*)data;

  if(incoming == NULL) return TRUE;
  if(incoming->socket == NULL) return TRUE;

  switch(condition) {
  case G_IO_IN:
    {
      GIOError error = G_IO_ERROR_NONE;
      guint bytes_read;

      error = g_io_channel_read(iochannel,
				&incoming->buffer[incoming->n],
				sizeof(incoming->buffer) - incoming->n,
				&bytes_read);

      if(error != G_IO_ERROR_NONE) {
	gchar *str = g_strdup_printf("%s: read error %d",__FUNCTION__,error);
	gnome_error_dialog(str);
	g_free(str);
      } else {
	if(bytes_read > 0) {
	  if(incoming->n == 0) {
	    g_io_add_watch(iochannel,G_IO_OUT,handling_iofunc,incoming);
	  }
	  incoming->n += bytes_read;
	}
      }
    } 
    break;
  case G_IO_OUT:
    {	
      if(incoming->buffer[0] != '\0') {
	if(parse_buffer(incoming,client_or_server) == FALSE) {
	  current_error_message_show();
	}
      }
      if(incoming != NULL) {
	memset(incoming->buffer,'\0',incoming->n);
	incoming->n = 0;
      }
    } 
    return FALSE;
  case G_IO_HUP:
  case G_IO_ERR:
  case G_IO_HUP | G_IO_ERR:
    {
      /* FIXME: add functions that deal with the effects of
       * a lost connection... depending on the type of client_or_server
       */
      connection_destroy(incoming);
    } 
    return FALSE;
  default:
    {
      gchar *str = g_strdup_printf("%s: unknown condition %d",
				   __FUNCTION__,condition);
      current_error_message_set(str);
      current_error_message_show();
      g_free(str);
    }
    break;
  }
  return TRUE;
  
}

gchar *assemble_network_string(const gchar *lib,const gchar *func,
			       gint argc,const gchar *argv[]) {

  GString *str = g_string_new(lib);
  gchar *rv = NULL,*tmp = NULL;
  gint i;

  str = g_string_append(str,SEPARATOR);
  str = g_string_append(str,func);

  str = g_string_append(str,SEPARATOR);
  str = g_string_append(str,tmp = g_strdup_printf("%d",argc));
  g_free(tmp);

  for(i=0;i<argc;i++) {
    str = g_string_append(str,SEPARATOR);
    str = g_string_append(str,argv[i]);    
  }

  /* there is no need for a special end of network string
   * delimiter, because argc is given, so you know how many arguments
   * there are to parse and therefor you know when to stop
   */
  str = g_string_append(str,SEPARATOR);

  rv = str->str;

  /* FALSE means that you have to free the character segment yourself,
   * that is the string "rv" is pointing to now...
   */
  g_string_free(str,FALSE);
  
  return rv;

}

gboolean parse_buffer(Connection *c,ClientOrServer *cos) {

  GHashTable *hash = cos->loaded_modules;
  gboolean rv = FALSE;
  gchar *lib = NULL, *func = NULL, **argv = NULL;
  gint argc,i;

  lib  = g_strdup(strtok(c->buffer,SEPARATOR));
  func = g_strdup(strtok(NULL,SEPARATOR));

  argc = atoi(strtok(NULL,SEPARATOR));
  /* The +2 because I sneakily pass two other arguments that
   * you can ignore when making plugins... the hidden arguments
   * are a pointer to the Connection and a pointer to the ClientOrServer
   */
  argv = (gchar**)g_malloc(sizeof(gchar*)*(argc+2));

  for(i=0;i<argc;i++) {
    argv[i] = g_strdup_printf("%s",strtok(NULL,SEPARATOR));
  }

  /* here's the sneaky stuff... hehehe watch out!!! 
   * the value of i is now argc, that means it points behind the
   * last argument... not to the last argument.
   */
  argv[i] = (gchar*)c;
  argv[i+1] = (gchar*)cos;

  g_print("%s: calling func %s %s %d",__FUNCTION__,lib,func,argc);
  rv = call_func(hash,lib,func,argc,(const gchar**)argv);

  g_free(lib);
  g_free(func);
  for(i=0;i<argc;i++) {
    g_free(argv[i]);
  }
  g_free(argv);

  return rv;

}

gint get_argc_from_argv(gchar *argv[]) {

  gchar *walker = argv[0];
  gint i = 0;

  while(walker != NULL) {
    i++;
    walker = argv[i];
  }

  return i;

}

gboolean call_remote_func(Connection *c,
			  const gchar *lib,const gchar *func,
			  gchar *argv[]) {

  gchar *buffer = NULL;
  GIOError error = G_IO_ERROR_NONE;
  gint bytes_written = 0;
  gboolean rv = TRUE;

  g_assert(c != NULL);

  buffer = assemble_network_string(lib,func,get_argc_from_argv(argv),
	                           (const gchar**)argv);
  g_print("--\nassembled string to send: [%s]\n--\n",buffer);
  sprintf(c->buffer,buffer);
  g_free(buffer);

  c->n = strlen(c->buffer);
  error = gnet_io_channel_writen(c->iochannel,c->buffer,c->n,&bytes_written);

  if(error != G_IO_ERROR_NONE) {
    gchar *str = g_strdup_printf(_("error contacting remote host %s"),c->ip);
    current_error_message_set(str);
    current_error_message_show();
    g_free(str);
    rv = FALSE;
  } 

  if(bytes_written == c->n) {
    c->n = 0;
    memset(c->buffer,'\0',SOCKET_BUFFER_SIZE);
  }

  return rv;

}
